from tkinter.font import Font 

def get_font(font):
    font_dict = {}
    if isinstance(font, (tuple,list)):
        for item in font:
            if item in ['normal', 'bold']:
                font_dict['weight'] = item 
            elif item in ['roman', 'italic']:
                font_dict['slant'] = item 
            elif isinstance(item, int):
                font_dict['size'] = item 
            elif isinstance(item, str) and item not in ['normal', 'bold', 'roman', 'italic']:
                font_dict['family'] = item
    else:
        font_dict['family'] = font
    try:
        return Font(**font_dict)
    except Exception as e:
        print("font error :", e)



from tkeditor.ui.line_number import LineNumber
from tkeditor.ui.autoscrollbar import AutoScrollbar
from tkeditor.ui.context_menu import ContextMenu
from tkeditor.ui.folding_code import FoldingCode
from tkeditor.ui.base_editor import CustomText


__all__ = ["LineNumber", "AutoScrollbar","ContextMenu","FoldingCode"]



from tkeditor.Lexers.baselexer import BaseLexer
import re

CSS_PROPERTIES = {
    "color", "background", "font-size", "font-family", "margin", "padding",
    "border", "width", "height", "display", "position", "top", "left",
    "right", "bottom", "flex", "grid", "align-items", "justify-content",
    "text-align", "line-height", "opacity", "overflow", "z-index"
}

class CSSLexer(BaseLexer):
    name = "css"
    file_extensions = ["css"]

    COMMENT_RE = re.compile(r'/\*.*?\*/', re.DOTALL)
    SELECTOR_RE = re.compile(r'^[^{]+(?=\{)')
    PROPERTY_RE = re.compile(r'([a-zA-Z-]+)\s*:')
    VALUE_RE = re.compile(r':\s*([^;]+);')
    OPERATOR_RE = re.compile(r'[{}:;]')

    def lex(self, text):
        tokens = []
        lines = text.split("\n")
        protected = []

        # 1) Comments (protected)
        for ln, line in enumerate(lines, start=1):
            for m in self.COMMENT_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("comment", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

        def is_protected(line_no, col):
            for pl, s, e in protected:
                if pl == line_no and s <= col < e:
                    return True
            return False

        # 2) Process each line
        for ln, line in enumerate(lines, start=1):

            # Selector
            m = self.SELECTOR_RE.match(line)
            if m:
                s, e = m.start(), m.end()
                if not is_protected(ln, s):
                    tokens.append(("selector", f"{ln}.{s}", f"{ln}.{e}"))

            # Properties
            for m in self.PROPERTY_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                prop = m.group(1)
                tokens.append(("property", f"{ln}.{s}", f"{ln}.{e}"))

            # Values
            for m in self.VALUE_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                tokens.append(("value", f"{ln}.{s}", f"{ln}.{e}"))

            # Operators
            for m in self.OPERATOR_RE.finditer(line):
                s, e = m.start(), m.end()
                if not is_protected(ln, s):
                    tokens.append(("operator", f"{ln}.{s}", f"{ln}.{e}"))

        return tokens



from tkinter import Frame
from tkeditor.core import CustomText
from tkeditor.ui import LineNumber, AutoScrollbar, FoldingCode, ContextMenu
from tkinter import ttk 
class Editor(Frame):
    def __init__(self, master, **kwarg):
        """
        A subclass of `tk.Text` that supports all standard `Text` parameters,
        plus additional options for enhanced code editing features.

        Additional Parameters:
            **scrollbg (str)**: Background color of the scrollbar trough.
            **thumbbg (str)**: Background color of the scrollbar thumb.
            **activescrollbg (str)**: Background color of the scrollbar thumb when active/hovered.
            **folder_code (bool)**: Enables or disables code folding.
            **linenumber (bool)**: Enables or disables the display of line numbers.
            **indent_line_color (str)**: Color of vertical indentation guide lines.
            **line_number_fg (str)**: Foreground (text) color of line numbers.
            **lineboxwidth (int)**: Width of the line number gutter in pixels.
            **folding_arrow_color (str)**: Color of the code-folding arrow indicator.
            **foldingboxwidth (int)**: Width of the folding gutter in pixels.
            **folding_bg (str)**: Background color of the folding gutter.
            **bracket_tracker_color (str)**: Highlight color for matching brackets.
            **bracket_tracker (bool)**: Highlighting brackets 
            **current_line_color (str)**: Background color for the current line highlight.
        """

        Allowed_keys = Frame(master).keys()
        super().__init__(master, **{k:v for k, v in kwarg.items() if k in Allowed_keys})
        self.master = master

        self.scrollbar_layout()
        

        self.line_number = LineNumber(self, **kwarg)
        self.folding_code = FoldingCode(self, **kwarg)
        self.text  = CustomText(self, **kwarg)
        self.v_scroll = AutoScrollbar(self, orient='vertical', command=self.text.yview, style='Custom.Vertical.TScrollbar')
        self.h_scroll = AutoScrollbar(self, orient='horizontal', command=self.text.xview, style='Custom.Horizontal.TScrollbar')

        self.text.config(yscrollcommand=self.v_scroll.set)
        self.text.config(xscrollcommand=self.h_scroll.set)


        self.line_number.attach(self.text)
        self.folding_code.attach(self.text)


        trough_color = kwarg.get('scrollbg') if kwarg.get('scrollbg') else self.text.cget('bg')
        thumb_color = kwarg.get('thumbbg') if kwarg.get('thumbbg') else "#5b5b5b"
        hover_color = kwarg.get('activescrollbg') if kwarg.get('activescrollbg') else self.text.cget('bg')
        self.create_scrollbar_style("Custom.Vertical.TScrollbar",trough_color=trough_color, thumb_color=thumb_color,hover_color=hover_color)
        self.create_scrollbar_style("Custom.Horizontal.TScrollbar",trough_color=trough_color, thumb_color=thumb_color,hover_color=hover_color)


        self.text.grid(row=0, column=2, sticky='nsew')
        self.v_scroll.grid(row=0, column=3, rowspan=2, sticky='ns')
        self.h_scroll.grid(row=1, column=2, sticky='ew')
        if kwarg.get('folding_code', True):
            self.folding_code.grid(row=0, column=1, rowspan=2, sticky='ns')
        if kwarg.get('linenumber',True):
            self.line_number.grid(row=0, column=0, rowspan=2, sticky='ns')

        self.rowconfigure(0, weight=1)
        self.columnconfigure(2, weight=1)
        
        
        self.context_menu = ContextMenu(self.text)
        self.context_menu.setup_context_menu()


        self.Events(**kwarg)
    
    def create_scrollbar_style(self,name: str, trough_color: str,thumb_color: str, hover_color: str):
        self.style.configure(name,
                        troughcolor=trough_color,
                        background=thumb_color,
                        bordercolor=trough_color,
                        lightcolor=thumb_color,
                        darkcolor=thumb_color,
                        arrowcolor=thumb_color,
                        gripcount=0)

        
        self.style.map(name, background=[('active', hover_color)])
    def scrollbar_layout(self):
        self.style = ttk.Style(self.master)
        self.style.theme_use('clam')
        self.style.layout('Custom.Vertical.TScrollbar', [
            ('Vertical.Scrollbar.trough', {
                'children': [
                    ('Vertical.Scrollbar.thumb', {'unit': '1', 'sticky': 'ns'})
                ],
                'sticky': 'ns'
            })
        ])
        self.style.layout('Custom.Horizontal.TScrollbar', [
            ('Horizontal.Scrollbar.trough', {
                'children': [
                    ('Horizontal.Scrollbar.thumb', {'unit': '1', 'sticky': 'ew'})
                ],
                'sticky': 'ew'
            })
        ])

    def debounce(self, widget, attr_name, delay, callback):
        after_id = getattr(widget, attr_name, None)
        if after_id:
            widget.after_cancel(after_id)
        setattr(widget, attr_name, widget.after(delay, callback))

    def Events(self, **kwarg):
        if kwarg.get("bracket_tracker", False):
            ### --- Bracket Tracker --- ###
            self.text.bind("<KeyRelease>", lambda e: self.debounce(self.text, "_track_bracket_after", 50, self.text.bracket_tracker.track_brackets), add="+")
            self.text.bind("<Button-1>", lambda e: self.text.after_idle(self.text.bracket_tracker.track_brackets), add="+")

        ### --- Folding Code --- ###
        for event in ("<Configure>", "<KeyRelease>", "<MouseWheel>", "<ButtonRelease-1>"):
            self.text.bind(event, self.folding_code._schedule_draw, add="+")
        self.v_scroll.bind("<B1-Motion>", self.folding_code._schedule_draw, add="+")
        self.folding_code.bind("<Button-1>", self.folding_code._on_click, add="+")

        ### --- Line Number --- ###
        
        for event in ("<KeyRelease>", "<MouseWheel>", "<Button-1>", "<Configure>","<<Redraw>>"):
            self.text.bind(event, self.line_number.schedule_redraw, add="+")
        self.v_scroll.bind("<B1-Motion>", self.line_number.schedule_redraw, add="+")

        ### --- Indentation Events --- ###
        self.text.bind("<Return>", self.text.indent.auto_indent, add="+")
        self.text.bind("<Control-Return>", self.text.indent.escape_line, add="+")
        self.text.bind("<BackSpace>", self.text.indent.backspace, add="+")

        ### --- Bracket Completion --- ###
        for key in ["[", "{", "(", "]", "}", ")", "'", '"']:
            self.text.bind(f"<Key-{key}>", self.text.brackets_and_string_complete, add="+")

        ### --- Tab Key --- ###
        self.text.bind("<Tab>", self.text._handle_tab, add="+")

        ### --- Current Line Highlight --- ###
        def highlight_line():
            self.debounce(self.text, "_linecolor_after", 50, self.text.set_current_line_color)

        self.text.bind("<Key>", lambda e: highlight_line(), add="+")
        self.text.bind("<Button-1>", lambda e: highlight_line(), add="+")
        self.text.bind("<B1-Motion>", lambda e: self.text.tag_remove('current_line', '1.0', 'end'), add="+")

        
        ### --- Optional Indentation Guide Trigger --- ###
        if kwarg.get('indentationguide', False):
            self.folding_code.bind("<Button-1>", lambda e: self.text.indentationguide.schedule_draw(), add="+")


    def configure(self, **kwarg):
        super().configure(**{k:v for k, v in kwarg.items() if k in Frame().keys()})
        if "linenumber" in kwarg.keys():
            if kwarg.get('linenumber'):
                self.line_number.grid(row=0, column=0,rowspan=2, sticky='ns')
            else:
                self.line_number.grid_remove()

        if "folding_code" in kwarg.keys():
            if kwarg.get('folding_code'):
                self.folding_code.grid(row=0, column=1, rowspan=2, sticky='ns')
            else:
                self.folding_code.grid_remove()

        self.text.configure(**kwarg)
        self.line_number.configure(**kwarg)
        self.folding_code.configure(**kwarg)

        if  'indent_line_color' in kwarg.keys():
            self.text.indentationguide.set_color(kwarg.get('indent_line_color', '#4b4b4b'))

        if 'folding_arrow_color' in kwarg.keys():
            self.folding_code.set_color(kwarg.get('folding_arrow_color', '#aaa'))

        if 'bracket_tracker_color' in kwarg.keys():
            self.text.bracket_tracker.set_color(kwarg.get('bracket_tracker_color','lightblue'))

        trough_color = kwarg.get('scrollbg') if kwarg.get('scrollbg') else self.text.cget('bg')
        thumb_color = kwarg.get('thumbbg') if kwarg.get('thumbbg') else "#5b5b5b"
        hover_color = kwarg.get('activescrollbg') if kwarg.get('activescrollbg') else self.text.cget('bg')
        self.create_scrollbar_style("Custom.Vertical.TScrollbar",
                                       trough_color=trough_color, 
                                       thumb_color=thumb_color,
                                       hover_color=hover_color
                                       )
        self.create_scrollbar_style("Custom.Horizontal.TScrollbar",
                                       trough_color=trough_color, 
                                       thumb_color=thumb_color,
                                       hover_color=hover_color
                                       )
    config = configure

    
    def bind(self, sequence=None, func=None, add=None):
        self.text.bind(sequence, func, add)
    def get_context_menu(self):
        """Return the context menu for the editor."""
        return self.text.context_menu.get_popup_menu()
    
    def get_content(self, condition:str):
        if condition.lower().strip() == "all":
            return self.text.get_all_content()
        elif condition.lower().strip() == "visible":
            return self.text.get_visible_content()
    def get_text_widget(self):
        return self.text


from tkinter import Text
from tkeditor.core.auto_indent import Indentations, IndentationGuide
from tkeditor.core.bracket_match import BracketTracker
class CustomText(Text):
    def __init__(self, master=None, **kwargs):
        Allowed_keys = Text(master).keys()
        super().__init__(master, **{k:v for k, v in kwargs.items() if k in Allowed_keys})
        self.tab_width = 4

        
        self.indent = Indentations(self)
        self.indentationguide = IndentationGuide(text=self, color=kwargs.get('indent_line_color','#4b4b4b'))    
        self.current_linecolor = kwargs.get('current_line_color', '#eee')
        self.tag_configure("current_line", background=self.current_linecolor)

        self.set_current_line_color()

        if kwargs.get('indentationguide',False):
            self.indentationguide.set_indentationguide()
        
        self.bracket_tracker = BracketTracker(self, kwargs.get('bracket_tracker_color','lightblue'))

        # self.Events()

    def Events(self):
        """Bind events to the text widget."""
        for key in ["[", "{", "(", "]", "}", ")", "'", '"']:
            super().bind(f"<Key-{key}>", self.brackets_and_string_complete, add="+")
        # super().bind('<Key>', self.brackets_and_string_complete, add="+")
        super().bind("<Tab>", self._handle_tab, add="+")
        super().bind("<Button-1>", self.set_current_line_color, add="+")
        super().bind("<Key>", self.set_current_line_color, add="+")
        super().bind("<B1-Motion>", lambda e: self.tag_remove('current_line','1.0','end'), add="+")
        
    def set_language(self, lang: str):
        self.indent.set_language(lang)

    def set_indentation(self, indent: int):
        self.indent.set_indentation(indent)

    
        
    def set_current_line_color(self, event=None):
        """Set the color for the current line in the editor."""
        def task():
            if not self.tag_ranges("sel"):
                self.tag_remove("current_line", "1.0", "end")
                self.tag_add("current_line", "insert linestart", "insert lineend+1c")
            else:
                self.tag_remove("current_line", "1.0", "end")
            self.tag_lower("current_line", "sel")
            self.tag_lower("current_line", "BracketTracker")

        self.after_idle(task)


    def configure(self, **kwargs):
        super().configure(**{k:v for k, v in kwargs.items() if k in Text().keys()})
        if "indentationguide" in kwargs.keys():
            if kwargs.get('indentationguide'):
                self.indentationguide.set_indentationguide()
            else:
                self.indentationguide.remove_indentationguide()

    def brackets_and_string_complete(self, event):
        char = event.char
        brackets = {"[":"]","(":")","{":"}","'":"'",'"':'"'}
        if char in brackets.keys():
            self.mark_gravity('insert','left')
            self.insert('insert', brackets[char])
            self.mark_gravity('insert','right')

    def set_tabWidth(self, width:int):
        """Set the tab width for the editor."""
        if isinstance(width, int) and width >= 0:
            if width > 0:
                self.tab_width = width
            else:
                raise ValueError("Tab width must be a positive integer.")
        else:
            raise TypeError("Tab width must be an integer.")
        
    def _handle_tab(self, event):
        """Handle tab key press for indentation."""
        self.insert("insert", " " * self.tab_width)
        self.set_current_line_color(event)
        return "break"

    def bind(self, sequence=None, func=None, add=None):
        if sequence == "<Key>":
            return super().bind(sequence, func, add="+")
        return super().bind(sequence, func, add=add)
    
    def get_all_content(self):
        return self.get("1.0", "end")
    
    def get_visible_content(self):
        first_line = self.index("@0,0")
        last_line = self.index(f"@0,{self.winfo_height()}")
        return self.get(first_line, last_line)

    


from tkeditor.Lexers.baselexer import BaseLexer
import re


JS_KEYWORDS = {
    "let","const","var",
    "function","return","class","extends","new",
    "if","else","switch","case","break","continue",
    "for","while","do",
    "import","from","export","default",
    "try","catch","finally","throw",
    "true","false","null","undefined",
}


class JavaScriptLexer(BaseLexer):

    name = "javascript"
    file_extensions = ["js", "mjs", "jsx"]

    STRING_RE = re.compile(r"""(["'`])((?:\\.|(?!\1).)*)\1""")
    IDENT_RE = re.compile(r"[A-Za-z_$][A-Za-z0-9_$]*")
    FUNC_CALL_RE = re.compile(r"([A-Za-z_$][A-Za-z0-9_$]*)\s*\(")
    CLASS_RE = re.compile(r"class\s+([A-Za-z_$][A-Za-z0-9_$]*)")

    def lex(self, text):

        tokens = []
        lines = text.split("\n")

        protected = []   # list of (line, start_col, end_col)

        # ---------------------------------
        # 1) Capture strings + comments first
        # ---------------------------------
        for ln, line in enumerate(lines, start=1):

            # ----- strings -----
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

            # ----- comments // -----
            if "//" in line:
                idx = line.index("//")
                tokens.append(("comment", f"{ln}.{idx}", f"{ln}.{len(line)}"))
                protected.append((ln, idx, len(line)))

        # ---------------------------------
        # helper: check if inside protected
        # ---------------------------------
        def is_protected(line, col):
            for pl, a, b in protected:
                if pl == line and a <= col < b:
                    return True
            return False

        # ---------------------------------
        # 2) Now identifiers, keywords, class, functions
        # ---------------------------------
        for ln, line in enumerate(lines, start=1):

            for m in self.IDENT_RE.finditer(line):
                s, e = m.start(), m.end()
                if is_protected(ln, s):
                    continue

                word = m.group(0)
                if word in JS_KEYWORDS:
                    tokens.append(("keyword", f"{ln}.{s}", f"{ln}.{e}"))
                else:
                    tokens.append(("ident", f"{ln}.{s}", f"{ln}.{e}"))

            # ----- class Foo -----
            for m in self.CLASS_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                if not is_protected(ln, s):
                    tokens.append(("class", f"{ln}.{s}", f"{ln}.{e}"))

            # ----- funcName( -----
            for m in self.FUNC_CALL_RE.finditer(line):
                fname = m.group(1)
                s, e = m.start(1), m.end(1)

                if is_protected(ln, s):  
                    continue
                if fname in JS_KEYWORDS:
                    continue

                tokens.append(("function", f"{ln}.{s}", f"{ln}.{e}"))

            # ----- operators -----
            for i, ch in enumerate(line):
                if ch in "{}[]()+-*/%=!<>:;,.":
                    if not is_protected(ln, i):
                        tokens.append(("operator", f"{ln}.{i}", f"{ln}.{i+1}"))

        return tokens



from tkeditor.Lexers.baselexer import BaseLexer
from tkeditor.Lexers.csslexer import CSSLexer
from tkeditor.Lexers.jslexer import JavaScriptLexer
import re

HTML_KEYWORDS = { "html","head","body","title","meta","link","script","style","div","span", ... }

class HTMLLexer(BaseLexer):
    name = "html"
    file_extensions = ["html","htm"]

    TAG_RE = re.compile(r"</?[A-Za-z0-9_\-]+")
    ATTR_RE = re.compile(r'\b([a-zA-Z\-:]+)\b(?=\=)')
    STRING_RE = re.compile(r'"[^"]*"|\'[^\']*\'')
    COMMENT_RE = re.compile(r"<!--.*?-->", re.DOTALL)
    OPERATOR_CHARS = set("<>/=:")

    def lex(self, text):
        tokens = []
        lines = text.split("\n")
        protected_lines = {}  # line_no -> list of (start, end)

        # -------------------------------
        # 1) Comments + Strings
        # -------------------------------
        for ln, line in enumerate(lines, start=1):
            protected_lines.setdefault(ln, [])
            for m in self.COMMENT_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("comment", f"{ln}.{s}", f"{ln}.{e}"))
                protected_lines[ln].append((s, e))
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected_lines[ln].append((s, e))

        # -------------------------------
        # Helper: fast protected check
        # -------------------------------
        def is_protected(line_no, col):
            for s, e in protected_lines.get(line_no, []):
                if s <= col < e:
                    return True
            return False

        # -------------------------------
        # 2) Parse line by line
        # -------------------------------
        ln = 0
        while ln < len(lines):
            line = lines[ln]
            # detect <script> or <style>
            match = re.search(r"<(script|style)(.*?)>", line, re.IGNORECASE)
            if match:
                tag = match.group(1).lower()
                start_tag_start = match.start()
                start_tag_end = match.end()

                # highlight opening tag
                tokens.append(("tag", f"{ln+1}.{start_tag_start}", f"{ln+1}.{start_tag_end}"))

                start_ln = ln
                # find closing tag
                end_ln = ln
                while end_ln < len(lines):
                    end_match = re.search(rf"</{tag}>", lines[end_ln], re.IGNORECASE)
                    if end_match:
                        closing_tag_start = end_match.start()
                        closing_tag_end = end_match.end()
                        break
                    end_ln += 1
                else:
                    closing_tag_start = closing_tag_end = 0

                # highlight closing tag
                tokens.append(("tag", f"{end_ln+1}.{closing_tag_start}", f"{end_ln+1}.{closing_tag_end}"))

                # inner content between tags
                inner = "\n".join(lines[start_ln+1:end_ln])
                if tag == "style":
                    nested_tokens = CSSLexer().lex(inner)
                else:
                    nested_tokens = JavaScriptLexer().lex(inner)

                # adjust line offsets for nested tokens...


                # adjust line offsets
                for tok_type, start, end in nested_tokens:
                    s_line, s_col = map(int, start.split("."))
                    e_line, e_col = map(int, end.split("."))
                    s_line += start_ln + 1
                    e_line += start_ln + 1
                    tokens.append((tok_type, f"{s_line}.{s_col}", f"{e_line}.{e_col}"))
                    # protect nested lines
                    for l in range(s_line, e_line+1):
                        protected_lines.setdefault(l, []).append((0, len(lines[l-1])))

                ln = end_ln  # skip processed block
            else:
                # normal tags, attrs, operators
                for m in self.TAG_RE.finditer(line):
                    s, e = m.start(), m.end()
                    if not is_protected(ln+1, s):
                        tokens.append(("tag", f"{ln+1}.{s}", f"{ln+1}.{e}"))
                for m in self.ATTR_RE.finditer(line):
                    s, e = m.start(1), m.end(1)
                    if not is_protected(ln+1, s):
                        tokens.append(("attribute", f"{ln+1}.{s}", f"{ln+1}.{e}"))
                for i, ch in enumerate(line):
                    if ch in self.OPERATOR_CHARS and not is_protected(ln+1, i):
                        tokens.append(("operator", f"{ln+1}.{i}", f"{ln+1}.{i+1}"))
            ln += 1

        return tokens



from tkeditor.Lexers.baselexer import BaseLexer
import re

CPP_KEYWORDS = {
    "auto","bool","break","case","catch","char","class","const","constexpr",
    "continue","decltype","default","delete","do","double","else","enum",
    "explicit","extern","false","float","for","friend","goto","if","inline",
    "int","long","mutable","namespace","new","noexcept","nullptr","operator",
    "private","protected","public","register","reinterpret_cast","return",
    "short","signed","sizeof","static","static_cast","struct","switch",
    "template","this","throw","true","try","typedef","typeid","typename",
    "union","unsigned","using","virtual","void","volatile","while"
}

class CPPLexer(BaseLexer):

    name = "cpp"
    file_extensions = ["cpp", "hpp", "cc", "cxx", "h"]

    STRING_RE   = re.compile(r'"(\\.|[^"])*"')
    CHAR_RE     = re.compile(r"'(\\.|[^'])'")
    IDENT_RE    = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
    FUNC_CALL_RE = re.compile(r"([A-Za-z_][A-Za-z0-9_]*)\s*\(")
    PREPROCESSOR_RE = re.compile(r"#\s*[A-Za-z_][A-Za-z0-9_]*")

    MULTI_OPS = [
        "==","!=", ">=","<=", "++","--", "&&","||",
        "<<", ">>", "+=","-=","*=","/=","%=",
        "::", "->", "->*", ".*"
    ]

    def lex(self, text):

        tokens = []
        lines = text.split("\n")

        protected = []   # Protected spans: strings, chars, comments

        # -------------------------------------------------------
        # 1) Strings, Chars, Single-line // Comments, Multi-line /* */
        # -------------------------------------------------------
        in_block_comment = False

        for ln, line in enumerate(lines, start=1):

            i = 0
            L = len(line)

            # MULTI-LINE COMMENTS /* ... */
            while i < L:
                if not in_block_comment and line.startswith("/*", i):
                    in_block_comment = True
                    start = i
                    i += 2
                    continue

                if in_block_comment:
                    endpos = line.find("*/", i)
                    if endpos == -1:
                        # continue comment to end of line
                        tokens.append(("comment", f"{ln}.{start}", f"{ln}.{L}"))
                        protected.append((ln, start, L))
                        break
                    else:
                        # comment ends on this line
                        tokens.append(("comment", f"{ln}.{start}", f"{ln}.{endpos+2}"))
                        protected.append((ln, start, endpos+2))
                        in_block_comment = False
                        i = endpos + 2
                        continue

                i += 1

            # STRING LITERALS
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

            # CHAR LITERALS
            for m in self.CHAR_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

            # SINGLE-LINE // COMMENTS
            if "//" in line:
                i = line.index("//")
                tokens.append(("comment", f"{ln}.{i}", f"{ln}.{len(line)}"))
                protected.append((ln, i, len(line)))
                
            m = self.PREPROCESSOR_RE.match(line)
            if m:
                s, e = m.start(), m.end()
                tokens.append(("preprocessor", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, len(line)))

        # -------------------------------------------------------
        # Helper: check if protected
        # -------------------------------------------------------
        def is_protected(line_no, col):
            for pl, a, b in protected:
                if pl == line_no and a <= col < b:
                    return True
            return False

        # -------------------------------------------------------
        # 2) IDENTIFIER, KEYWORD, FUNCTION CALL, OPERATORS
        # -------------------------------------------------------
        for ln, line in enumerate(lines, start=1):

            # IDENTIFIERS + KEYWORDS
            for m in self.IDENT_RE.finditer(line):
                s, e = m.start(), m.end()
                if is_protected(ln, s):
                    continue

                word = m.group(0)

                if word in CPP_KEYWORDS:
                    tokens.append(("keyword", f"{ln}.{s}", f"{ln}.{e}"))
                else:
                    tokens.append(("ident", f"{ln}.{s}", f"{ln}.{e}"))

            # FUNCTION CALLS: name(
            for m in self.FUNC_CALL_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                name = m.group(1)

                if is_protected(ln, s):
                    continue
                if name in CPP_KEYWORDS:
                    continue

                tokens.append(("function", f"{ln}.{s}", f"{ln}.{e}"))

            # MULTI-CHAR OPERATORS
            for op in self.MULTI_OPS:
                idx = line.find(op)
                while idx != -1:
                    if not is_protected(ln, idx):
                        tokens.append(("operator", f"{ln}.{idx}", f"{ln}.{idx+len(op)}"))
                    idx = line.find(op, idx + 1)

            # SINGLE CHARACTER OPERATORS
            for i, ch in enumerate(line):
                if ch in "{}[]()+-*/%=!<>:;,.&|^~?":
                    if not is_protected(ln, i):
                        tokens.append(("operator", f"{ln}.{i}", f"{ln}.{i+1}"))

        return tokens



from tkeditor.Lexers.baselexer import BaseLexer
import re


C_KEYWORDS = {
    "int","char","float","double","void","long","short",
    "for","while","do","if","else","switch","case","return","break","continue",
    "struct","typedef","enum","union","sizeof","static","const","volatile",
    "signed","unsigned","goto",
}


class CLexer(BaseLexer):

    name = "c"
    file_extensions = ["c", "h"]

    STRING_RE = re.compile(r'"(\\.|[^"])*"')
    CHAR_RE = re.compile(r"'(\\.|[^'])'")
    IDENT_RE = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
    FUNC_CALL_RE = re.compile(r"([A-Za-z_][A-Za-z0-9_]*)\s*\(")
    PREPROCESSOR_RE = re.compile(r"#\s*[A-Za-z_][A-Za-z0-9_]*")

    def lex(self, text):

        tokens = []
        lines = text.split("\n")

        protected = []   # (line, start, end)

        # -----------------------------------
        # 1) Strings + comments (protected)
        # -----------------------------------
        for ln, line in enumerate(lines, start=1):

            # string literals
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

            # char literals
            for m in self.CHAR_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))
            
            # preprocessor directives
            m = self.PREPROCESSOR_RE.match(line)
            if m:
                s, e = m.start(), m.end()
                tokens.append(("preprocessor", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, len(line)))

            # comments //
            if "//" in line:
                i = line.index("//")
                tokens.append(("comment", f"{ln}.{i}", f"{ln}.{len(line)}"))
                protected.append((ln, i, len(line)))
            
        # -----------------------------------
        # helper: check if inside protected
        # -----------------------------------
        def is_protected(line_no, col):
            for pl, a, b in protected:
                if pl == line_no and a <= col < b:
                    return True
            return False
        #numbers
        for m in re.finditer(r'\b\d+(\.\d+)?\b', line):
            s, e = m.start(), m.end()
            if not is_protected(ln, s):
                tokens.append(("number", f"{ln}.{s}", f"{ln}.{e}"))
        # -----------------------------------
        # 2) IDENT, KEYWORD, FUNCTION, OPERATOR
        # -----------------------------------
        for ln, line in enumerate(lines, start=1):

            # identifiers + keywords
            for m in self.IDENT_RE.finditer(line):
                s, e = m.start(), m.end()

                if is_protected(ln, s):
                    continue

                word = m.group(0)

                if word in C_KEYWORDS:
                    tokens.append(("keyword", f"{ln}.{s}", f"{ln}.{e}"))
                else:
                    tokens.append(("ident", f"{ln}.{s}", f"{ln}.{e}"))

            # function call: name(
            for m in self.FUNC_CALL_RE.finditer(line):
                fname = m.group(1)
                s, e = m.start(1), m.end(1)

                if is_protected(ln, s):
                    continue

                if fname in C_KEYWORDS:
                    continue

                tokens.append(("function", f"{ln}.{s}", f"{ln}.{e}"))

            # operators
            for i, ch in enumerate(line):
                if ch in "{}[]()+-*/%=!<>:;,.&|^~?":
                    if not is_protected(ln, i):
                        tokens.append(("operator", f"{ln}.{i}", f"{ln}.{i+1}"))

        return tokens



from tkinter import Frame, Text
from tkinter.ttk import Treeview, Style
from tkeditor.utils.helper import get_font
import re

WORD_RE = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")

class DropDown(Frame):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.editor: Text = master.get_text_widget()
        self.style = Style(self)

        self.font = get_font(self.editor.cget("font"))
        self.style.configure("Custom.Treeview", font=self.font)

        self.box = Treeview(self, show="tree", selectmode="browse", style="Custom.Treeview")
        self.box.pack(expand=True, fill="both")


        self.editor.bind("<KeyRelease>", self.show_dropdown, add="+")
        self.box.bind("<Double-1>", self.insert_selection)


    def collect_words(self) -> set:
        """Collect unique words typed in editor"""
       
        return  self.editor.lexer.identifiers


    def show_dropdown(self, event):
        # ignore control keys
        if not event.char or not event.char.isprintable():
            self.place_forget()
            return

        # get current prefix
        index = self.editor.index("insert")
        line = self.editor.get(index + " linestart", index)
        match = re.findall(r"[A-Za-z_][A-Za-z0-9_]*$", line)
        if not match:
            self.place_forget()
            return

        prefix = match[0]

        # collect words
        words = self.collect_words()
        suggestions = sorted([w for w in words if w.startswith(prefix) and w != prefix])

        if not suggestions:
            self.place_forget()
            return

        # display popup
        self.box.delete(*self.box.get_children())
        for w in suggestions:
            self.box.insert("", "end", text=w)

        bbox = self.editor.bbox("insert")
        if not bbox:
            self.place_forget()
            return

        x, y, _, _ = bbox
        self.place_configure(
            x=x, 
            y=(y + self.font.metrics("linespace")), 
            height=self.font.metrics("linespace")*len(suggestions) if len(suggestions) <= 6 else self.font.metrics("linespace")*6
        )
        self.lift()


    def insert_selection(self, event=None):
        selection = self.box.focus()
        if not selection:
            return
        word = self.box.item(selection, "text")

        # delete the current prefix
        index = self.editor.index("insert")
        line = self.editor.get(index + " linestart", index)
        prefix = re.findall(r"[A-Za-z_][A-Za-z0-9_]*$", line)[0]
        self.editor.delete(f"insert-{len(prefix)}c", "insert")

        # insert chosen completion
        self.editor.insert("insert", word)
        self.place_forget()



from tkinter import Canvas, TclError
from tkeditor.utils import get_font
class FoldingCode(Canvas):
    def __init__(self, master=None, **kwargs):
        Allowed_keys = Canvas(master).keys()
        super().__init__(master, **{k: v for k, v in kwargs.items() if k in Allowed_keys})
        self.configure(
            width=kwargs.get('foldingboxwidth', 20),
            bg=kwargs.get('folding_bg', self.cget('background') or self.cget('bg')),
            highlightthickness=0
        )
        self.text_widget = None
        self.folded_blocks = {}  
        self.tag_prefix = "folded_"
        self.font = get_font(kwargs.get('font', ('Consolas', 14)))
        self.fg = kwargs.get('folding_arrow_color', kwargs.get('fg', '#000000'))
        self.width = kwargs.get('width')
        
    def set_color(self, color):
        """Set the color for folding code."""
        self.fg = color
        self.itemconfig("folding_line", fill=color)
    def attach(self, text_widget):
        self.text_widget = text_widget
        # for event in ("<Configure>", "<KeyRelease>", "<MouseWheel>", "<ButtonRelease-1>"):
        #     self.text_widget.bind(event, self._schedule_draw, add="+")
        # self.bind("<Button-1>", self._on_click)
        self._schedule_draw()

    def _schedule_draw(self, event=None):
        if not hasattr(self, "_draw_scheduled") or not self._draw_scheduled:
            self._draw_scheduled = True
            self.after_idle(self._draw_folding_lines)

    def _draw_folding_lines(self):
        self._draw_scheduled = False
        self.delete("folding_line")
        if not self.text_widget:
            return

        index = self.text_widget.index("@0,0")
        while True:
            dline = self.text_widget.dlineinfo(index)
            if dline is None:
                break

            y = dline[1]
            lineno = index.split(".")[0]
            line_no_int = int(lineno)
            is_hidden = False
            for start, end in self.folded_blocks.items():
                if int(start) < line_no_int < int(end):
                    is_hidden = True
                    break
            if is_hidden:
                index = self.text_widget.index(f"{index}+1line")
                continue

            line_text = self.text_widget.get(f"{lineno}.0", f"{lineno}.end")

            x = 5  # Gutter margin

            # Detect foldable lines (you can expand this)
            
            if line_text.strip() and (line_text.strip().endswith(":") or self._has_block(int(lineno))):
                folded = lineno in self.folded_blocks
                try:
                    # test if font supports arrow
                    symbol = "❯" if folded else "∨"
                except:
                    symbol = "+" if folded else "-"
                self.create_text(x, y, text=symbol, anchor="nw", font=self.font, fill=self.fg, tags=("folding_line", f"line_{lineno}"))

            index = self.text_widget.index(f"{index}+1line")
        # Adjust gutter width if needed
        required_width = 1 * self.font.measure("M") + 10
        if required_width != self.width:
            self.width = required_width
            self.config(width=self.width)
    def _has_block(self, line_number):
        """Return True if the next line is more indented than this one"""
        current_line = self.text_widget.get(f"{line_number}.0", f"{line_number}.end")
        base_indent = self._get_indent(current_line)

        # Check next few lines for indentation
        for offset in range(1, 10):  # Max 10 lines lookahead
            try:
                next_line = self.text_widget.get(f"{line_number + offset}.0", f"{line_number + offset}.end")
            except TclError:
                return False
            if not next_line.strip():
                continue
            if self._get_indent(next_line) > base_indent:
                return True
            else:
                return False
        return False

    def _on_click(self, event):
        clicked = self.find_withtag("current")
        if not clicked:
            return
        tags = self.gettags(clicked[0])
        line_tag = next((tag for tag in tags if tag.startswith("line_")), None)
        if not line_tag:
            return
        lineno = line_tag.replace("line_", "")
        if lineno in self.folded_blocks:
            self._unfold_block(lineno)
        else:
            self._fold_block(lineno)

    def _fold_block(self, start_line):
        start_index = f"{start_line}.0"
        end_index = self._find_block_end(start_index)
        if not end_index:
            return

        tag = self.tag_prefix + start_line
        self.text_widget.tag_add(tag, f"{start_index}+1line", end_index)
        self.text_widget.tag_configure(tag, elide=True)
        self.folded_blocks[start_line] = end_index.split(".")[0]
        self.text_widget.event_generate("<<Redraw>>")  
        self._schedule_draw()

    def _unfold_block(self, start_line):
        tag = self.tag_prefix + start_line
        self.text_widget.tag_remove(tag, f"{start_line}.0", f"{self.folded_blocks[start_line]}.end")
        self.folded_blocks.pop(start_line, None)
        self.text_widget.event_generate("<<Redraw>>")  
        self._schedule_draw()

    def _find_block_end(self, start_index):
        lines = self.text_widget.get(start_index, "end-1c").splitlines()
        start_line_no = int(start_index.split(".")[0])
        base_indent = self._get_indent(lines[0])
        for i, line in enumerate(lines[1:], start=1):
            if line.strip() == "":
                continue
            indent = self._get_indent(line)
            if indent <= base_indent:
                return f"{start_line_no + i}.0"
        return f"{start_line_no + len(lines)}.0"

    def _get_indent(self, line):
        return len(line) - len(line.lstrip(" "))



from tkeditor.Lexers.baselexer import BaseLexer
import re

class JSONLexer(BaseLexer):
    name = "json"
    file_extensions = ["json"]

    STRING_RE = re.compile(r'"(\\.|[^"])*"')
    NUMBER_RE = re.compile(r'\b-?(0|[1-9]\d*)(\.\d+)?([eE][+-]?\d+)?\b')
    BOOLEAN_RE = re.compile(r'\b(true|false|null)\b')

    def lex(self, text):
        tokens = []
        lines = text.split("\n")
        protected = []  # [(line_no, start_col, end_col)]

        # 1) Strings (protected)
        for ln, line in enumerate(lines, start=1):
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

        def is_protected(line_no, col):
            for pl, s, e in protected:
                if pl == line_no and s <= col < e:
                    return True
            return False

        # 2) Numbers, booleans, braces, colons, commas
        for ln, line in enumerate(lines, start=1):
            for m in self.NUMBER_RE.finditer(line):
                s, e = m.start(), m.end()
                if not is_protected(ln, s):
                    tokens.append(("number", f"{ln}.{s}", f"{ln}.{e}"))

            for m in self.BOOLEAN_RE.finditer(line):
                s, e = m.start(), m.end()
                if not is_protected(ln, s):
                    tokens.append(("keyword", f"{ln}.{s}", f"{ln}.{e}"))  # boolean/null as keyword

            # operators: { } [ ] : ,
            for i, ch in enumerate(line):
                if ch in "{}[]:," and not is_protected(ln, i):
                    tokens.append(("operator", f"{ln}.{i}", f"{ln}.{i+1}"))

        return tokens



from tkeditor.ui.base_editor import CustomText
from tkeditor.core.manage import debounce, run_in_thread, throttle, timeit, cached_measure_indent
from tkeditor.core.auto_indent import Indentations, IndentationGuide
__all__ = [
    "Indentations", 
    "IndentationGuide", 
    "CustomText",
    "debounce",
    "run_in_thread",
    "throttle",
    "timeit",
    "cached_measure_indent"
]


from tkinter import Canvas
from tkeditor.utils import get_font

class LineNumber(Canvas):
    def __init__(self, master, **kwarg):
        Allowed_keys = Canvas(master).keys()
        super().__init__(master, **{k:v for k, v in kwarg.items() if k in Allowed_keys})
        self.master = master
        self.kwarg = kwarg
        self.text_widget = None
        self.fill = kwarg.get('line_number_fg','black')
        self.font = get_font(kwarg.get('font',("TkDefaultFont",9)))
        self.width = kwarg.get('lineboxwidth', 55)
        self.current_line_number = kwarg.get('current_line_number', "steelblue")

        self.char_width = self.font.measure('M')

        self.config(width=self.width, 
                    bg=kwarg.get("line_number_bg") if kwarg.get("line_number_bg") else kwarg.get('bg') or kwarg.get("background")
                    )
        self.fill = kwarg.get("line_number_fg") if kwarg.get("line_number_fg") else kwarg.get("fg") or kwarg.get("foreground")
    def schedule_redraw(self,event=None):
        self.text_widget.after_idle(self.refresh_lines)
    def attach(self, text_widget):
        self.text_widget = text_widget

        
    def refresh_lines(self, event=None):
        self.redraw()

    def redraw(self):
        self.delete('all')
        index = self.text_widget.index("@0,0")
        max_digits = 0
        char_width = self.char_width
        y_to_line = {}  # <-- map y coordinate to the LAST line at that y

        while True:
            dline = self.text_widget.dlineinfo(index)
            if not dline:
                break  # No more visible lines
            
            y = dline[1]
            lineno = index.split('.')[0]
            y_to_line[y] = lineno  # <-- overwrite so last line at y wins
            index = self.text_widget.index(f"{index}+1line")

        for y in sorted(y_to_line):
            lineno = y_to_line[y]
            color = self.fill
            if lineno == self.text_widget.index("insert").split('.')[0]:
                color = self.current_line_number
            max_digits = max(max_digits, len(lineno))
            x = self.width - len(lineno) * char_width - 5
            self.create_text(x, y, anchor='nw', text=lineno,
                             font=self.font, fill=color,
                             tags=('line_number'))

        # Adjust gutter width if needed
        required_width = max_digits * char_width + 10
        if required_width != self.width:
            self.width = required_width
            self.config(width=self.width)



    def configure(self, **kwarg):
        super().configure(**kwarg)
        if "line_number_fg" in kwarg:
            self.itemconfigure("line_number", fill=self.kwarg.get("line_number_fg"))



from tkinter import Frame
import re

class Indentations:
    def __init__(self, text_widget):
        self.text_widget = text_widget
        # self.setup_auto_indent()
        self.indentation = 4
        self.language = "python"  # default
        
    def setup_auto_indent(self):
        self.text_widget.bind("<Return>", self.auto_indent, add="+")
        self.text_widget.bind("<Control-Return>", self.escape_line, add="+")
        self.text_widget.bind("<BackSpace>", self.backspace, add="+")
        self.language = "python"  # default

    def set_language(self, lang: str):
        self.language = lang.lower()

    def set_indentation(self, indent: int):
        if isinstance(indent, int) and indent > 0:
            self.indentation = indent
        else:
            raise ValueError("Indentation must be a positive integer.")

    def auto_indent(self, event):
        current_line = self.text_widget.get("insert linestart", "insert")
        match = re.match(r"(\s*)", current_line)
        if match:
            indent = len(match.group(1))
            line_text = current_line.strip()

            # Language-specific block starters
            if self.language == "python":
                if line_text.endswith(":"):
                    indent += self.indentation
                

            elif self.language in ("html", "xml"):
                if re.match(r"<[^/!][^>]*[^/]?>$", line_text):  
                    indent += self.indentation

            elif self.language == "lua":
                if re.search(r"\b(then|do|function)\b$", line_text):
                    indent += self.indentation

            elif self.language == "yaml":
                if line_text.endswith(":"):
                    indent += self.indentation
            if line_text.endswith("{") or line_text.endswith("(") or line_text.endswith("["):
                current_indent = indent
                indent += self.indentation
                self.text_widget.insert("insert", "\n" + " " * indent)
                self.text_widget.mark_gravity('insert', 'left')
                self.text_widget.insert("insert", "\n" + " " * current_indent)
                self.text_widget.mark_gravity('insert', 'right')
            else:
                self.text_widget.insert("insert", "\n" + " " * indent)
            self.text_widget.see("insert")
            self.text_widget.event_generate("<<Redraw>>")
            self.text_widget.event_generate("<<DropDown>>")
            self.text_widget.set_current_line_color(event)
            return "break"
        self.see("insert")
    def escape_line(self, event):
        current_line = self.text_widget.get("insert linestart", "insert lineend")
        match = re.match(r"(\s*)", current_line)
        if not match:
            return

        indent = len(match.group(1))
        line_text = current_line.strip()
        language = self.language.lower()

        # Block starters and indent rules
        increase_indent = False
        if language == "python":
            increase_indent = line_text.endswith(":") or ("{" in line_text and not "}" in line_text)
        elif language in ("c", "cpp", "java", "javascript", "csharp"):
            increase_indent = "{" in line_text and not line_text.startswith("}")
        elif language in ("html", "xml"):
            increase_indent = re.match(r"<[^/!][^>]*[^/]?>$", line_text) is not None
        elif language == "lua":
            increase_indent = re.search(r"\b(then|do|function)\b$", line_text) is not None
        elif language == "yaml":
            increase_indent = line_text.endswith(":")

        if increase_indent:
            indent += self.indentation

        # Clean insert logic
        line_content = self.text_widget.get("insert", "insert lineend")
        self.text_widget.delete("insert", "insert lineend")
        self.text_widget.insert("insert", line_content + "\n" + " " * indent)
        self.text_widget.see("insert")
        self.text_widget.set_current_line_color(event)
        self.text_widget.event_generate("<<Redraw>>")
        self.text_widget.event_generate("<<DropDown>>")

        return "break"


    def backspace(self, event):
        current_line = self.text_widget.get("insert linestart", "insert")
        if current_line.isspace() and not self.text_widget.tag_ranges("sel"):
            if len(current_line) % self.indentation == 0:
                self.text_widget.delete(f"insert-{self.indentation}c", "insert")
                self.text_widget.event_generate("<<Redraw>>")
                self.text_widget.event_generate("<<DropDown>>")
                self.text_widget.set_current_line_color(event)
                return "break"
        self.text_widget.event_generate("<<Redraw>>")
        self.text_widget.event_generate("<<DropDown>>")

        self.text_widget.set_current_line_color(event)


class IndentationGuide:
    def __init__(self, text, color=None):
        self.text = text
        self.color = color if color else '#4b4b4b'
        self.indent_lines = []
    def set_color(self, color):
        """Set the color for indentation guide."""
        self.color = color
        for frame in self.indent_lines:
            frame.config(background=color)
    # def set_indentationguide(self):
    #     self.text.original_yview = self.text.yview
    #     self.text.yview = self.yview_wrapper
    #     self.text.original_xview = self.text.xview
    #     self.text.xview = self.xview_wrapper
    #     self._indent_guide_binds = []
    #     self._indent_guide_binds.append(self.text.bind("<KeyRelease>", self.schedule_draw, add="+"))
    #     self._indent_guide_binds.append(self.text.bind("<MouseWheel>", self.schedule_draw, add="+"))
    #     self._indent_guide_binds.append(self.text.bind("<Configure>", self.schedule_draw, add="+"))
    #     self._indent_guide_binds.append(self.text.bind("<ButtonRelease-1>", self.schedule_draw, add="+"))

    # def remove_indentationguide(self):
    #     self.text.original_yview = self.text.yview
    #     self.text.original_xview = self.text.xview
    #     events = ["<KeyRelease>", "<MouseWheel>", "<Configure>", "<ButtonRelease-1>"]
    #     for event, bind_id in zip(events, getattr(self, "_indent_guide_binds", [])):
    #         if bind_id:
    #             self.text.unbind(event, bind_id)
    #     self._indent_guide_binds = []
    def debounce(self,widget, attr_name, delay, callback):
        after_id = getattr(widget, attr_name, None)
        if after_id:
            widget.after_cancel(after_id)
        setattr(widget, attr_name, widget.after(delay, callback))

    def set_indentationguide(self):
        if getattr(self, "_indent_guide_enabled", False):
            return
        self._indent_guide_enabled = True

        if not hasattr(self.text, "original_yview"):
            self.text.original_yview = self.text.yview
            self.text.original_xview = self.text.xview

        self.text.yview = self.yview_wrapper
        self.text.xview = self.xview_wrapper
        self._indent_guide_binds = []

        def draw(): self.schedule_draw()

        def debounced_draw(delay):
            def inner(event=None):
                self.debounce(self.text, "_indent_draw_after", delay, draw)
            return inner

        for event in ("<KeyRelease>", "<Configure>", "<ButtonRelease-1>"):
            bind_id = self.text.bind(event, debounced_draw(100), add="+")
            self._indent_guide_binds.append((event, bind_id))

        # Immediate update on scroll
        mw_id = self.text.bind("<MouseWheel>", debounced_draw(0), add="+")
        self._indent_guide_binds.append(("<MouseWheel>", mw_id))

    def remove_indentationguide(self):
        if not getattr(self, "_indent_guide_enabled", False):
            return  # already removed

        self._indent_guide_enabled = False

        # Restore original view methods
        if hasattr(self.text, "original_yview"):
            self.text.yview = self.text.original_yview
        if hasattr(self.text, "original_xview"):
            self.text.xview = self.text.original_xview

        for event, bind_id in getattr(self, "_indent_guide_binds", []):
            if bind_id:
                self.text.unbind(event, bind_id)
        self._indent_guide_binds = []

        # Clean up debounced job
        after_id = getattr(self, "_indent_draw_after", None)
        if after_id:
            self.text.after_cancel(after_id)
            del self._indent_draw_after


    def schedule_draw(self, event=None):
        self.text.after_idle(self.draw_indent)

    def yview_wrapper(self, *args):
        self.text.original_yview(*args)
        self.schedule_draw()

    def xview_wrapper(self, *args):
        self.text.original_xview(*args)
        self.schedule_draw()

    def draw_indent(self):
        for frame in self.indent_lines:
            frame._used = False

        first = self.text.index("@0,0")
        first_index = int(first.split('.')[0])
        last = self.text.index(f"@0,{self.text.winfo_height()} +1line")
        visible_text = self.text.get(first, last)

        frame_index = 0
        prev_indent_levels = 0
        for line_no, line in enumerate(visible_text.splitlines(), start=first_index):
            if line.strip() == "":
                indent_levels = prev_indent_levels  # Use previous indent
            else:
                match = re.match(r'^\s+', line)
                if match:
                    indent = match.group()
                    indent_width = 4
                    indent_levels = len(indent.replace('\t', ' ' * indent_width)) // indent_width
                    prev_indent_levels = indent_levels
                else:
                    indent_levels = 0
                    prev_indent_levels = 0

            for i in range(indent_levels):
                char_index = f"{line_no}.{i * 4}"
                bbox = self.text.bbox(char_index)
                if bbox:
                    x, y, width, height = bbox
                    if frame_index < len(self.indent_lines):
                        frame = self.indent_lines[frame_index]
                    else:
                        frame = Frame(self.text, background=self.color)
                        self.indent_lines.append(frame)
                    frame.place(x=x, y=y, width=2, height=height)
                    frame._used = True
                    frame_index += 1

        for frame in self.indent_lines[frame_index:]:
            frame.place_forget()



from tkinter import ttk, TclError
class AutoScrollbar(ttk.Scrollbar):
    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
        ttk.Scrollbar.set(self, lo, hi)
    def pack(self):
        return TclError('Cannot use pack with this widget')
    def place(self):
        return TclError('Cannot use place with this widget')
    


from tkeditor.Lexers.baselexer import BaseLexer
import keyword
import tokenize, io

PYTHON_KEYWORDS = set(keyword.kwlist + [
    "self", "cls"])


class PythonLexer(BaseLexer):

    name = "python"
    file_extensions = ["py"]
    def __init__(self):
        super().__init__()
        self.identifiers = set()
    def lex(self, text):

        tokens = []
        src = io.StringIO(text)

        prev_tok = None
        prev_tok_is_decorator = False  # initialize here!
        in_fstring = False
        store_word = []
        

        try:
            for tok in tokenize.generate_tokens(src.readline):

                ttype = tok.type
                tstr  = tok.string

                sline, scol = tok.start
                eline, ecol = tok.end

                start = f"{sline}.{scol}"
                end   = f"{eline}.{ecol}"

                # ====================================================
                # 1) BASIC TOKENS
                # ====================================================
                # -------------------------------
                # f-string handling
                # -------------------------------
                if ttype == tokenize.FSTRING_START:
                    tokens.append(("string", start, end))
                    in_fstring = True

                elif ttype == tokenize.FSTRING_MIDDLE:
                    # The literal parts between {…} are string
                    # Expressions inside {} are separate tokens
                    if "{" in tstr or "}" in tstr:
                        tokens.append(("f_expr", start, end))
                    else:
                        tokens.append(("string", start, end))

                elif ttype == tokenize.FSTRING_END:
                    tokens.append(("string", start, end))
                    in_fstring = False
                    
                if ttype == tokenize.STRING:
                    tokens.append(("string", start, end))
                elif ttype == tokenize.COMMENT:
                    tokens.append(("comment", start, end))

                elif ttype == tokenize.NUMBER:
                    tokens.append(("number", start, end))

                elif tstr in PYTHON_KEYWORDS:
                    tokens.append(("keyword", start, end))

                elif ttype == tokenize.OP:
                    if tstr == "@":
                        prev_tok_is_decorator = True
                        tokens.append(("operator", start, end))
                    else:
                        tokens.append(("operator", start, end))

                elif ttype == tokenize.NAME:
                    
                    if prev_tok_is_decorator:
                        tokens.append(("decorator", f"{sline}.{scol}", f"{eline}.{ecol}"))
                        prev_tok_is_decorator = False
                    else:
                        tokens.append(("ident", start, end))
                        
                        
                        

                
                # ====================================================
                # 2) CLASS NAME DETECTION: class NAME
                # ====================================================
                if (
                    prev_tok and
                    prev_tok.string == "class" and
                    ttype == tokenize.NAME
                ):
                    tokens.append(("class", start, end))


                # ====================================================
                # 3) FUNCTION CALL DETECTION: name (
                #    FIXED: do NOT mark keywords as functions
                #    FIXED: do NOT mark inside strings/comments
                # ====================================================
                if (
                    prev_tok and
                    prev_tok.type == tokenize.NAME and
                    prev_tok.string not in PYTHON_KEYWORDS and  # IMPORTANT FIX
                    ttype == tokenize.OP and
                    tstr == "("
                ):
                    # mark prev token as function
                    # fname_start = f"{prev_tok.start[0]}.{prev_tok.start[1]}"
                    # fname_end   = f"{prev_tok.end[0]}.{prev_tok.end[1]}"
                    # tokens.append(("function", fname_start, fname_end))
                    # skip if prev_tok is decorator (@name)
                    if prev_tok.start[0] == tok.start[0]:  # same line as '('
                        tokens.append(("function", f"{prev_tok.start[0]}.{prev_tok.start[1]}", f"{prev_tok.end[0]}.{prev_tok.end[1]}"))

                prev_tok = tok

        except tokenize.TokenError:
            pass

        return tokens




from tkeditor.Lexers.baselexer import BaseLexer
import re

class XMLLexer(BaseLexer):
    name = "xml"
    file_extensions = ["xml", "xhtml", "svg"]

    # Patterns
    TAG_RE = re.compile(r'</?([A-Za-z0-9:_-]+)')
    ATTR_RE = re.compile(r'([A-Za-z0-9:_-]+)\s*=')
    STRING_RE = re.compile(r'"(\\.|[^"])*"|\'(\\.|[^\'])*\'')
    COMMENT_RE = re.compile(r'<!--(.*?)-->')

    def set_style(self, style:dict) -> dict:
        self.style = style

    def lex(self, text):
        tokens = []
        lines = text.split("\n")
        protected = []  # (line, start, end)

        # 1) Comments (protected)
        for ln, line in enumerate(lines, start=1):
            for m in self.COMMENT_RE.finditer(line):
                s, e = m.start(), m.end()
                tokens.append(("comment", f"{ln}.{s}", f"{ln}.{e}"))
                protected.append((ln, s, e))

        def is_protected(line_no, col):
            for pl, s, e in protected:
                if pl == line_no and s <= col < e:
                    return True
            return False

        # 2) Tags, attributes, strings
        for ln, line in enumerate(lines, start=1):
            # strings
            for m in self.STRING_RE.finditer(line):
                s, e = m.start(), m.end()
                if not is_protected(ln, s):
                    tokens.append(("string", f"{ln}.{s}", f"{ln}.{e}"))
                    protected.append((ln, s, e))

            # tags
            for m in self.TAG_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                if not is_protected(ln, s):
                    tokens.append(("tag", f"{ln}.{s}", f"{ln}.{e}"))

            # attributes
            for m in self.ATTR_RE.finditer(line):
                s, e = m.start(1), m.end(1)
                if not is_protected(ln, s):
                    tokens.append(("attribute", f"{ln}.{s}", f"{ln}.{e}"))

            # operators: < > / = 
            for i, ch in enumerate(line):
                if ch in "<>/=" and not is_protected(ln, i):
                    tokens.append(("operator", f"{ln}.{i}", f"{ln}.{i+1}"))

        return tokens



import time
import threading
from functools import wraps, lru_cache


def debounce(wait_ms=100):
    """
    Debounce decorator: schedules function to run after wait_ms of inactivity.
    Useful for scroll/key/mouse events in Tkinter.
    """
    def decorator(func):
        timer_attr = f"_{func.__name__}_debounce_timer"

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            widget = getattr(self, 'text', None)
            if widget is None:
                return func(self, *args, **kwargs)
            if hasattr(self, timer_attr):
                widget.after_cancel(getattr(self, timer_attr))
            timer_id = widget.after(wait_ms, lambda: func(self, *args, **kwargs))
            setattr(self, timer_attr, timer_id)

        return wrapper
    return decorator



def throttle(wait_ms=100):
    """
    Throttle decorator: ensures function is called at most once every wait_ms.
    Useful for limiting heavy UI updates.
    """
    def decorator(func):
        last_call_attr = f"_{func.__name__}_last_call"

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            now = time.time() * 1000  # convert to ms
            last_call = getattr(self, last_call_attr, 0)
            if now - last_call >= wait_ms:
                setattr(self, last_call_attr, now)
                return func(self, *args, **kwargs)

        return wrapper
    return decorator


def timeit(func):
    """
    Timeit decorator: prints how long the function took to execute.
    Good for debugging performance bottlenecks.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        end = time.perf_counter()
        print(f"{func.__name__} took {end - start:.4f} seconds")
        return result

    return wrapper


@lru_cache(maxsize=256)
def cached_measure_indent(font_str: str) -> int:
    """
    Cache example: measure indent width (only for pure functions).
    Use this pattern for any non-GUI repeated heavy calls.
    """
    import tkinter.font as tkfont
    return tkfont.Font(font=font_str).measure("    ")

def run_in_thread(func):
    """
    Run function in a separate thread.
    Useful for long-running tasks that shouldn't block the UI.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True).start()
    return wrapper



import tokenize
import io
class BracketTracker:
    def __init__(self, text_widget, color=None):
        self.text_widget = text_widget
        self.open_brackets = {
            '(': ')',
            '{': '}',
            '[': ']'
        }
        self.close_brackets = {
            ')': '(',
            '}': '{',
            ']': '['
        }
        bgcolor = color if color else 'lightblue'
        self.text_widget.tag_configure("BracketTracker", background = bgcolor)
        # self.text_widget.bind("<KeyRelease>", lambda e:self.track_brackets(), add="+")
        # self.text_widget.bind("<Button-1>", lambda e: self.text_widget.after_idle(self.track_brackets), add="+")
    def set_color(self, color):
        """Set the color for bracket tracking."""
        self.text_widget.tag_configure("BracketTracker", background = color)
    def track_brackets(self):
        self.text_widget.tag_remove("BracketTracker", "1.0", "end")

        cursor_index = self.text_widget.index("insert")
        line = self.text_widget.get("insert linestart", "insert lineend")
        col = int(cursor_index.split(".")[1])

        # Do not stop here — we only skip brackets *inside* strings/comments
        pre_word = self.text_widget.get("insert -1c", "insert")
        after_word = self.text_widget.get("insert", "insert +1c")

        try:
            if pre_word in self.close_brackets.keys():
                self._match_backwards(pre_word)
            elif after_word in self.open_brackets.keys():
                self._match_forwards(after_word)
            else:
                # Handle case where cursor is BETWEEN matching ()
                # like:   (|)  ← cursor is between
                prev = self.text_widget.get("insert -1c", "insert")
                next = self.text_widget.get("insert", "insert +1c")
                if prev in self.open_brackets.keys() and next == self.open_brackets[prev]:
                    # Check that neither bracket is inside string/comment
                    line = self.text_widget.get("insert linestart", "insert lineend")
                    col = int(self.text_widget.index("insert").split(".")[1])
                    if not self.is_inside_string_or_comment(line, col - 1) and not self.is_inside_string_or_comment(line, col):
                        self.text_widget.tag_add("BracketTracker", "insert -1c", "insert")
                        self.text_widget.tag_add("BracketTracker", "insert", "insert +1c")
                else:
                    nearest = None  # (index, close_bracket)

                    # Search backwards to find the closest closing bracket
                    for close_bracket in self.close_brackets:
                        index = self._match_backwards(close_bracket, return_only=True)
                        if index:
                            if (nearest is None) or (self.text_widget.compare(index, ">", nearest[0])):
                                nearest = (index, close_bracket)

                    if nearest:
                        open_index, close_char = nearest
                        open_char = self.close_brackets[close_char]

                        self.text_widget.mark_set("match_temp", open_index)
                        close_index = self._match_forwards(open_char, return_only=True, start_index="match_temp +1c")

                        if close_index:
                            self.text_widget.tag_add("BracketTracker", open_index, f"{open_index} +1c")
                            self.text_widget.tag_add("BracketTracker", close_index, f"{close_index} +1c")
        except Exception as e:
            print("Bracket track error:", e)

    def _match_backwards(self, close_char, return_only=False, start_index=None):
        open_char = self.close_brackets.get(close_char)
        current_index = start_index or self.text_widget.index("insert -1c")

        depth = 1
        while True:
            # Search backwards for either char
            idx = self.text_widget.search(f"[{close_char}{open_char}]", current_index, stopindex="1.0", backwards=True, regexp=True)
            if not idx:
                return None

            ch = self.text_widget.get(idx, f"{idx} +1c")
            if ch == close_char:
                depth += 1
            elif ch == open_char:
                depth -= 1

            if depth == 0:
                if return_only:
                    return idx
                self.text_widget.tag_add("BracketTracker", idx, f"{idx} +1c")
                self.text_widget.tag_add("BracketTracker", self.text_widget.index("insert -1c"), "insert")
                return

            current_index = idx

    def _match_forwards(self, open_char, return_only=False, start_index=None):
        close_char = self.open_brackets.get(open_char)
        current_index = start_index or self.text_widget.index("insert +1c")

        depth = 1
        while True:
            # Search forwards for either char
            idx = self.text_widget.search(f"[{open_char}{close_char}]", current_index, stopindex="end-1c", forwards=True, regexp=True)
            if not idx:
                return None

            ch = self.text_widget.get(idx, f"{idx} +1c")
            if ch == open_char:
                depth += 1
            elif ch == close_char:
                depth -= 1

            if depth == 0:
                if return_only:
                    return idx
                self.text_widget.tag_add("BracketTracker", idx, f"{idx} +1c")
                self.text_widget.tag_add("BracketTracker", self.text_widget.index("insert"), "insert +1c")
                return

            current_index = f"{idx} +1c"


    def is_inside_string_or_comment(self,code: str, target_index: int) -> bool:
        try:
            tokens = tokenize.generate_tokens(io.StringIO(code).readline)
            for tok_type, tok_string, (start_line, start_col), (end_line, end_col), _ in tokens:
                if tok_type in (tokenize.STRING, tokenize.COMMENT):
                    if start_col <= target_index < end_col:
                        return True
        except Exception as e:
            pass
        return False


from tkinter import Menu

class ContextMenu:
    def __init__(self, text):
        self.text = text
        self.popup = Menu(self.text, tearoff=0)
        
        self.popup.add_command(label="Cut", command=self._cut)
        self.popup.add_command(label="Copy", command=self._copy)
        self.popup.add_command(label="Paste", command=self._paste)
        
        # self.setup_context_menu()

    def _cut(self):
        self.text.event_generate("<<Cut>>")
        self.popup.unpost()

    def _copy(self):
        self.text.event_generate("<<Copy>>")
        self.popup.unpost()

    def _paste(self):
        self.text.event_generate("<<Paste>>")
        self.popup.unpost()

    def setup_context_menu(self):
        self.text.bind("<Button-3>", self.popup_menu, add="+")

    def popup_menu(self, event):
        self.popup.post(event.x_root, event.y_root)

    def get_popup_menu(self):
        return self.popup



class BaseLexer:
    """Base class for all lexers."""
    name = "base"
    file_extensions = []   # ["py"], ["js"], etc.
    def __init__(self):

        
        self.default_styles = {
            # ---------- Core ----------
            "ident":        {"foreground":"#000000"},  # default text / variables
            "keyword":      {"foreground":"#ff4791"},  # bright pink / for keywords like if, else, return
            "function":     {"foreground":"#4d79ff"},  # function names
            "class":        {"foreground":"#ad6cff"},  # class names
            "operator":     {"foreground":"#ff8c42"},  # + - * / = etc
            "number":       {"foreground":"#f9c74f"},  # numeric literals
            "string":       {"foreground":"#00ff9c"},  # string literals
            "f_expr":       {"foreground":"#00ff9c"},  # string literals
            "comment":      {"foreground":"#9ca0a6"},  # comments, gray
            "decorator":    {"foreground":"#de5602"},
            # ---------- HTML / XML ----------
            "tag":          {"foreground":"#ff6e6e"},  # <html>, <div> etc
            "attribute":    {"foreground":"#f4a261"},  # class, id, src
            "value":        {"foreground":"#00ff9c"},  # attribute values (strings)
            
            # ---------- CSS ----------
            "property":     {"foreground":"#ffb347"},  # color, width, font-size
            "selector":     {"foreground":"#ad6cff"},  # .class, #id, tag selectors
            "unit":         {"foreground":"#f72585"},  # px, em, %
            
            # ---------- C / C++ ----------
            "preprocessor": {"foreground":"#ff6f61"},  # #include, #define
            
        }
    
    def set_styles(self, styles:dict):
        self.styles = styles

    def get_styles(self):
        if hasattr(self, "styles"):
            return self.styles
        else:
            return self.default_styles


    def lex(self, text):
        """
        MUST return list of tuples:
        [
            ("token_type", "start_index", "end_index"),
            ...
        ]
        Example: ("keyword", "1.0", "1.3")
        """
        raise NotImplementedError



from tkeditor.Lexers.baselexer import BaseLexer
from tkeditor.Lexers.pythonlexer import PythonLexer
from tkeditor.Lexers.jslexer import JavaScriptLexer
from tkeditor.Lexers.highlight import Highlighter
from tkeditor.Lexers.clexer import CLexer
from tkeditor.Lexers.cpplexer import CPPLexer
from tkeditor.Lexers.htmllexer import HTMLLexer
from tkeditor.Lexers.csslexer import CSSLexer
from tkeditor.Lexers.jsonlexer import JSONLexer
from tkeditor.Lexers.yamllexer import YAMLLexer
from tkeditor.Lexers.xmllexer import XMLLexer
__all__ = [
    "BaseLexer",
    "PythonLexer",
    "JavaScriptLexer",
    "Highlighter",
    "CLexer",
    "CPPLexer",
    "HTMLLexer",
    "CSSLexer",
    "JSONLexer",
    "YAMLLexer",
    "XMLLexer",
]


class Highlighter:
    def __init__(self, text_widget, lexer):
        self.text = text_widget
        self.lexer = lexer

        setattr(text_widget, 'lexer', self.lexer)

        # create tags
        self.styles = lexer.get_styles()
        for name, attrs in self.styles.items():
            self.text.tag_configure(name, **attrs)

        self.text.bind("<<Modified>>", self.schedule, add="+")
        self.after_id = None

    def schedule(self, event=None):
        if self.text.edit_modified():
            self.text.edit_modified(False)
            if self.after_id:
                self.text.after_cancel(self.after_id)
            self.after_id = self.text.after(60, self.highlight)

    def highlight(self):
        code = self.text.get("1.0", "end-1c")

        # remove all tags
        for name in self.styles:
            self.text.tag_remove(name, "1.0", "end")

        # apply new tokens
        for token_type, start, end in self.lexer.lex(code):
            self.text.tag_add(token_type, start, end)
